#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#define EMPTY     0
#define BUSY      1
#define SERVICING 2
#define LEAVING   3

struct BarberShop {
    pthread_mutex_t m;
    
    pthread_cond_t barber[2];
    pthread_cond_t w_room;
    pthread_cond_t customer_service[2];
    pthread_cond_t barber_end[2];
    
    int state[2];
    int wcustomers;
};

typedef struct BarberShop Shop;

void shop_init(Shop *s);
void shop_wait_for_customers(Shop *s, int seat);
int shop_wait_for_seat(Shop *s);
void shop_free_seat(Shop *s, int seat);
void shop_wait_for_service(Shop *s, int seat);


void shop_init(Shop *s)
{
    pthread_mutex_init(&s->m, 0);
    pthread_cond_init(&s->w_room, 0);
    pthread_cond_init(&s->barber[0], 0);
    pthread_cond_init(&s->barber[1], 0);
    pthread_cond_init(&s->customer_service[0], 0);
    pthread_cond_init(&s->customer_service[1], 0);
    pthread_cond_init(&s->barber_end[0], 0);
    pthread_cond_init(&s->barber_end[1], 0);
    s->state[0] = s->state[1] = EMPTY;
    s->wcustomers = 0;
}


void shop_wait_for_customers(Shop *s, int seat)
{
    pthread_mutex_lock(&s->m);

    while (s->state[seat] != BUSY) 
        pthread_cond_wait(&s->barber[seat], &s->m);

    s->state[seat] = SERVICING;
    pthread_cond_signal(&s->customer_service[seat]);

    pthread_mutex_unlock(&s->m);
}

int shop_wait_for_seat(Shop *s)
{
    int seat; 
    pthread_mutex_lock(&s->m);

    if (s->wcustomers == 5) {
        pthread_mutex_unlock(&s->m);
        return -1;
    }
    s->wcustomers++;
    while (s->state[0] != EMPTY && s->state[1] != EMPTY) { 
        pthread_cond_wait(&s->w_room, &s->m);        
    }
    s->wcustomers--;
    
    if (s->state[0] == EMPTY) seat = 0;
    else seat = 1;

    s->state[seat] = BUSY;

    pthread_cond_signal(&s->barber[seat]);
    pthread_mutex_unlock(&s->m);
    return seat;
}

void shop_free_seat(Shop *s, int seat)
{
    pthread_mutex_lock(&s->m);

    while (s->state[seat] != LEAVING)
        pthread_cond_wait(&s->barber_end[seat], &s->m);
    
    s->state[seat] = EMPTY;

    pthread_cond_signal(&s->w_room);
    
    pthread_mutex_unlock(&s->m);
}

void shop_wait_for_service(Shop *s, int seat)
{
    pthread_mutex_lock(&s->m);

    while (s->state[seat] != SERVICING) 
        pthread_cond_wait(&s->customer_service[seat], &s->m);
    
    s->state[seat] = LEAVING;
    pthread_cond_signal(&s->barber_end[seat]);

    pthread_mutex_unlock(&s->m);
}

Shop myshop;

void *barber(void *arg)
{
    int index = *((int*)arg);
    
    while(1) {
        printf("Barber   %d: ready\n", index);
        shop_wait_for_customers(&myshop, index);        
        printf("Barber   %d: serving a customer\n", index);
        sleep(3);
        printf("Barber   %d: end of service\n", index); 
        shop_free_seat(&myshop, index);
    }
}

void *customer(void *arg)
{
    int k = *((int*) arg);
    
    printf("Customer %d: enters the barbershop\n", k);
    int seat = shop_wait_for_seat(&myshop);
    if (seat == -1) {
        printf("Customer %d: barbershop full, exiting\n", k);
        return 0;
    }
    else {
        printf("Customer %d: being served,\n", k);
        shop_wait_for_service(&myshop, seat);
        printf("Customer %d: finished, exiting\n", k);
    }
    return 0;
}


int main(int argc, char *argv[])
{
    pthread_t *cust_tid;
    int *cust_index;
    pthread_t barbers[2];
    int barbers_id[2];
    int i;
    
    if (argc < 2) {
        printf("Usage: %s <ncustomers>\n", argv[0]);
        exit(0);
    }
    
    int ncust = atoi(argv[1]);

    shop_init(&myshop);
    
    printf("creating %d customers\n", ncust);
 
    cust_tid = (pthread_t *)malloc(sizeof(pthread_t)*ncust);
    cust_index = (int *)malloc(sizeof(int)*ncust);
    
    for(i=0; i<2; i++) {
        barbers_id[i] = i;
        pthread_create(&barbers[i], 0, barber, &barbers_id[i]);
    }
    
    for (i=0; i<ncust; i++) {
        cust_index[i] = i;
        pthread_create(&cust_tid[i], 0, customer, &cust_index[i]);
        sleep(1);
    }
          
    for (i=0; i<ncust; i++) {
        pthread_join(cust_tid[i], 0);
    }
            
    printf("FINISHED!\n");
    return 0;
}
